
	.INCLUDE inc\global.inc
	;.INCLUDE inc\dirp00.inc
	.INCLUDE inc\dpcbmfs.inc

;-command errors--------------------
badsyn		.EQU	$30   ;  general syntax
badcmd		.EQU	$31   ;  invalid command
longln		.EQU	$32   ;  long line
badfn		.EQU	$33   ;  invalid filname
nofile		.EQU	$34   ;  no file given
nocfil		.EQU	$39   ;  comnd file not found

	.CODEFD
	.LONGA off
	.LONGI off

e_nofile:
	lda	#nofile		; no files
	bra	cmderr
e_badsynw:
	;sta	!erword		; **could be warning
e_badsyn:
	lda	#badsyn		; err: bad syntax	
cmderr:
	stz	track		; cmd level err proc
	stz	sector
	jmp	$0000

; tag cmd string,setup cmd.. struc, image & file pntrs
; parse command line for R,S,N
;
; image bits
;            pgdrpgdr
;            fs1 fs2
;   bit reps:  pattern
;              greater than one file 
;              default drive(s) 
;              required filename

tagcmd:	jsr	prscln
	beq	e_nofile	; missing ':' -- error

tagcmd2:
	dey
	dey
	sty	filtbl		; ":"-1 starts fs1
	txa
	bne	e_nofile	; err: "," before ":"

prsequ:
	lda	#'='		; search: "="
	jsr	parse
	txa			; ?file count= 1-1?
	beq	?40
	lda	#%01000000	; g1-bit - fs1 count > 1
?40:	ora	#%00100001	; d1,r2-bits - def. drive fs1, fs2 file req.
	sta	!image		; fs structure
	inx
	stx	f1cnt		; fs1 file count
	stx	f2cnt		; init for no fs2
	lda	!patflg
	beq	?50
	lda	#%10000000	; p1-bit
	tsb	image
	stz	patflg		; clear pattern flag
?50:	tya			; ptr to fs2
	beq	?75		; no '=' -> no fs2
	cpx	#mxfils
	bcs	?55		; file table full ?
	sta	!filtbl,x
?55:	lda	!f1cnt		; fs2 is here now,...
	sta	!f2ptr		; ...now set f2 ptr
	lda	#$8d		; dummy: find cr-shifted (till end of line)
	jsr	parse		; parse rest of cmd string
	inx			; advance filtbl ptr to end
	stx	f2cnt		; save it
	dex			; restore for test
	lda	!patflg		; save last pattern
	beq	?60		; ?any patterns?
	lda	#%00001000	; yes, p2-bit
?60:	cpx	f1cnt		; ?f2cnt=f1cnt+1?
	beq	?70
	ora	#%00000100	; g2-bit
?70:	ora	#%00000011	; d2-bit,r2-bit
	eor	!image		; eor clears r2-bit
	sta	!image
?75:	lda	!image
	ldx	cmdnum
	and	>struct,x	; match cmd template
	bne	e_badsynw
	rts

; simple parser - set just the drive number
simprs:	stz	filtbl
	jsr	prscln		; search colon
	beq	?02		; ":" not found, Y = next char
	dey
	dey
	sty	filtbl		; ptr to char before ":"
?02:	stz	image
	ldy	filtbl
?04:	lda	cmdbuf,y
	jsr	tst0v1
	bpl	?20		; ok, found drive #
	iny
	cpy	cmdsiz
	bcs	?10		; end of string
	ldy	cmdsiz
	dey
	bne	?04
?10:	dec	image		; invalidate cmd template
	lda	lstdrv		; use last accessed drive
?20:	and	#1
	sta	drvnum
	rts

;test char in accum for "0" or "1"
tst0v1:	cmp	#'0'
	beq	?02
	cmp	#'1'
	beq	?02
	ora	#$80		; use default drive
?02:	and	#$81		; mask dr #
	rts      

; find pos'n of ":"
prscln:	ldy	#0		; buffer ptr
	tyx			; comma counter
	lda	#':'         

;parse string
;  looks for special chars,
;  returning when var'bl char
;  is found
;   a: var'bl char
;   x: in,out: index, filtbl+1
;   y: in: index, cmdbuf
;     out: new ptr, =0 if none
;         (z=1) if y=0
parse:	sta	!char		; save var'bl char
?10:	lda	cmdbuf,y	; match char
	beq	?30		; end of string
	iny
	cmp	!char
	beq	?35		; found char
	cmp	#'*'		; match pattern chars
	beq	?20
	cmp	#'?'
	bne	?25
?20:	inc	!patflg		; set pattern flag
?25:	cmp	#','		; match file separator
	bne	?10		; loop
	cpx	#mxfils-1
	bcs	?26		; no store if table full
	tya
	sta	!filtbl+1,x	; put ptrs in table
?26:	lda	!patflg		; save pattern for file
	and	#$7f
	beq	?28
	lda	#$80		; retain pattern presence...
	sta	pattyp,x
	sta	!patflg		; ...but clear count
?28:	inx
	cpx	#mxfils-1
	bcc	?10		; no more than mxfils
?30:	ldy	#0		; y=0 (z=1)
?35:	cpx	#mxfils-1
	bcs	?37		; no store if table full
	lda	!cmdsiz
	sta	!filtbl+1,x
?37:	lda	!patflg
	and	#$7f
	beq	?40
	lda	#$80
	sta	pattyp,x
?40:	tya			; z is set
	rts

; set 1st drive and table pointers
onedrv:	lda	!f2cnt		; save f2cnt
	sta	!f1cnt
	lda	#1		; set pointer to first file
	sta	!f2cnt
	sta	!f2ptr
	
;set up all drives from f2cnt (all files in stream)
alldrs:	ldy	lstdrv 
	ldx	#0		; ...into file entry table...
?10	stx	f1ptr		; ...on sector ptr byte
	lda	!filtbl,x
	jsr	setdrv
	ldx	f1ptr
	sta	!filtbl,x	; incr ptr past ":"
	tya			; bits rep drives
	sta	!fildrv,x	; bit7: default
	inx			; bit0: drive #
	cpx	f2cnt
	bcc	?10
	rts      

;---------------------------------------
;  set drive number
;  determines drive # from text or
;  uses default (-d)
;  a: in,out: index, cmdbuf
;  y: in: default drive
;    out: drive number, - if default
;---------------------------------------
setdrv:	tax			; x= cmdbuf index
	lda	#':'
	cmp	!cmdbuf+1,x	; for xxx:file
	beq	?40		;       ^
	cmp	!cmdbuf,x	; for xxx:file
	bne	?50		;        ^
	inx			; found ":", so...
?20:	tya			; drive= default
?22:	and	#1		; convert to numeric
?24:	tay			; restore drive
	txa			; a=index & xxxxfile
	rts			;               ^
?40:	lda	!cmdbuf,x
	inx			;  xxx:file
	inx			;    --^
	cmp	#'0'		; for xx0:file
	beq	?22		;         ^
	cmp	#'1'		; for xx1:file
	beq	?22		;         ^
	bne	?20		; cmd:file or xx,:file
				;   ^           ^
?50:	tya			; for xxx,file or xx=file
	ora	#$80		;         ^          ^
	and	#$81		; drive= -default
	bne	?24		; finish testing

struct:
	.DB	%11011101	; rename
	.DB	%00011100	; scratch
	.DB	%10011110	; new
